import GPLinearODEMaker.powers_of_negative_one

"""
    mp_kernel(hyperparameters, δ, dorder; shift_ind=0)

Created by kernel_coder(). Requires 3 hyperparameters.
Likely created using mp_kernel_base() as an input.
Use with include("src/kernels/mp_kernel.jl").

# Arguments
- `hyperparameters::Vector`: The hyperparameter values. For this kernel, they should be `["mp_P", "m_λ", "p_amp"]`
- `δ::Real`: The difference between the inputs (e.g. `t1 - t2`)
- `dorder::Vector{<:Integer}`: How many times to differentiate with respect to the inputs and the `hyperparameters` (e.g. `dorder=[0, 1, 0, 2]` would correspond to differentiating once w.r.t the second input and twice w.r.t `hyperparameters[2]`)
- `shift_ind::Integer=0`: If changed, the index of which hyperparameter is the `δ` shifting one
"""
function mp_kernel(
    hyperparameters::AbstractVector{<:Real},
    δ::Real,
    dorder::AbstractVector{<:Integer})

    @assert length(hyperparameters)==3 "hyperparameters is the wrong length"
    dorder_len = 5
    @assert length(dorder)==dorder_len "dorder is the wrong length"
    @assert maximum(dorder) < 5 "No more than two derivatives for each time or hyperparameter can be calculated"
    @assert minimum(dorder) >= 0 "No integrals"

    dorder2 = dorder[2]
    dorder[2] += dorder[1]

    dorder_view = view(dorder, 2:dorder_len)
    
    mp_P = hyperparameters[1]
    m_λ = hyperparameters[2]
    p_amp = hyperparameters[3]

    dabs_δ = powers_of_negative_one(δ < 0)  # store derivative of abs()
    abs_δ = abs(δ)

    if dorder_view==[6, 0, 0, 2]
        func = (128*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 128*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 4*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 300.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^2 + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*π^4*cos(δ*π/mp_P)^2/mp_P^4)*dabs_δ^2/m_λ^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-32*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 16*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 200.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2/m_λ^2 + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(128*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 128*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^6)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*p_amp^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 800.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 480*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 480*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 360*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 - 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 2400.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 2400.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 1440*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 720*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 46080*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 116736*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^6 + 40960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^6*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^6*cos(δ*π/mp_P)^2/mp_P^6 + 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 + 122880*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^5 - 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^5 - 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 1920*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 1920*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 15360*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 3840*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 7680*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 + 122880*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^5 - 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^5 - 128000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 204800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^4*dabs_δ^2*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 76800*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 38400*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 122880*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^4 + 61440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^4 + 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 19200*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 6400*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 - 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 38400*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 61440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^4 - 19200*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^4*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 5760*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 5760*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 1920*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 3840*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 7680*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 2880*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 960*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 2880*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 64000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 51200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 38400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^2*m_λ^2) + 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 19200*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^4*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 38400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^2*m_λ^2) + 23040*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 11520*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 - 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 11520*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 5760*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P - 9600.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 2880*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 38400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[5, 0, 0, 2]
        func = 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 80*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 133.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 + 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 80*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 240*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 80*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 240*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 533.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 - 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 800.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 240*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 + 20480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^5 - 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^5 - 12800*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 20480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^4 + 1066.66666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 320*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 640*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1280*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 12800*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 20480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^4 - 10666.6666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 8533.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*dabs_δ^2*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 6400*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 3200*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 3200*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^4*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P - 1600.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 960*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^2*m_λ^2) + 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[4, 0, 0, 2]
        func = 6*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 4*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2/m_λ^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 16*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 160.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 96*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 48*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 48*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 4096*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^4 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 192*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 2560.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^2*m_λ^2) + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 - 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 192*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 2560.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[3, 0, 0, 2]
        func = 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)/mp_P^3 + 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 48*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 640.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P + 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[2, 0, 0, 2]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 13.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 8*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 53.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*dabs_δ^2*sin(δ*π/mp_P)^4/m_λ^2 + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^2 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[1, 0, 0, 2]
        func = -4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[0, 0, 0, 2]
        func = -4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4
    end

    if dorder_view==[6, 0, 1, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(128*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 128*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^6)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 60*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 45*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*dabs_δ^2/m_λ^3 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 4*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 1341.64078649987*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^4 - 600.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^3 - 600.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2/m_λ^3 + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 90*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(128*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 128*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^6)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 111.80339887499*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*abs_δ*dabs_δ^2/m_λ^4 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*p_amp*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 33.5410196624969*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 400.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 670.820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^4 + 134.164078649987*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 402.492235949962*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 670.820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*dabs_δ^2/m_λ^4 + 402.492235949962*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 100.623058987491*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 1788.85438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^3*sin(δ*π/mp_P)^2/m_λ^4 + 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 8.94427190999916*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 447.213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 268.328157299975*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 402.492235949962*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 201.246117974981*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 402.492235949962*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 13312*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^5*π^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^6 - 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 - 5366.5631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3*sin(δ*π/mp_P)^2/m_λ^4 + 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 - 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*p_amp*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 268.328157299975*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 268.328157299975*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 960*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 + 51200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^3) - 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 14310.8350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 28621.6701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*dabs_δ^3*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 2683.28157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 1609.96894379985*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 960*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 29766.5369164772*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^2) - 22897.3360895978*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^5*abs_δ*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^2) - 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) - 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 3434.60041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 3434.60041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) - 57243.3402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^4) - 34346.0041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 25600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 10733.126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^3) + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 10733.126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 536.65631459995*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 536.65631459995*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 2146.6252583998*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 3434.60041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) - 34346.0041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 17173.0020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 34346.0041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 14310.8350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 28621.6701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 17173.0020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 3577.70876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 17173.0020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*abs_δ*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 9600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^3) - 1609.96894379985*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 1609.96894379985*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 536.65631459995*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 2146.6252583998*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 10733.126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 3219.9378875997*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 1073.3126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 6439.87577519939*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 3219.9378875997*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 1609.96894379985*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[5, 0, 1, 1]
        func = 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 223.606797749979*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^3/m_λ^4 + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 10*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 200.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^3 + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 223.606797749979*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^4 + 67.0820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 22.3606797749979*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 894.427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^3*sin(δ*π/mp_P)^2/m_λ^4 + 266.666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 894.427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*dabs_δ^3*sin(δ*π/mp_P)^2/m_λ^4 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 44.7213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 44.7213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 67.0820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 298.142396999972*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 178.885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 44.7213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 160*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 2133.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 4266.66666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 894.427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) - 5724.33402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 5724.33402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 2385.13917599978*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 4770.27835199955*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 3577.70876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^3) + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^3) - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 89.4427190999916*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 357.770876399966*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 5724.33402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 1788.85438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 357.770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 1073.3126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 357.770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 1788.85438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 1073.3126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 1073.3126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 1073.3126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[4, 0, 1, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 40.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^3 + 12*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 6*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 6*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 4*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 44.7213595499958*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2/m_λ^4 + 26.8328157299975*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 26.8328157299975*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 357.770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3*sin(δ*π/mp_P)^2/m_λ^4 + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 8.94427190999916*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 178.885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 512*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 178.885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 1144.86680447989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 640.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^3) - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 214.66252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 71.5541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 429.32505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 214.66252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 214.66252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 214.66252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 214.66252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[3, 0, 1, 1]
        func = 3*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 3*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 6.70820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^3*sin(δ*π/mp_P)^2/m_λ^4 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 + 6.70820393249937*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^3) - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 71.5541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 143.108350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 178.885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^4) - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[2, 0, 1, 1]
        func = (8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 2.23606797749979*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.6666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^3 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 29.8142396999972*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^4 - 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[1, 0, 1, 1]
        func = -4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[0, 0, 1, 1]
        func = -4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/m_λ^2
    end

    if dorder_view==[6, 1, 0, 1]
        func = 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(640*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 128*δ*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 128*δ*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^7) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-768*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 768*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^7 - 512*δ*p_amp*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^8)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 60*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(128*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 128*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 128*δ*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(640*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 128*δ*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 128*δ*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^7) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(128*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 128*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 128*δ*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(128*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 128*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 128*δ*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(128*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 128*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 128*δ*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 300.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^2 + 60*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-384*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 384*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7 - 256*δ*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^8)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 600.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2/m_λ^2 - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^2/m_λ^2 - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 73728*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^8 - 40960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^7*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^5*cos(δ*π/mp_P)^3/mp_P^8 - 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^6*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^6*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^7 - 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^2/m_λ^2 + 8*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 16*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 960*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 960*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 3840*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P - 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^6*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^6*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^7 - 102400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^5*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/(mp_P^6*m_λ^2) - 61440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^6 - 30720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^6 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 - 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(128*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 128*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^6)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 30720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^5*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^6 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^5*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 2880*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ*p_amp^3*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 960*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 400.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 800.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 480*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 3840*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 240*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 51200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) - 51200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) - 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^3*m_λ^2) + 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 2880*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^5*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 5760*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 - 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 2880*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 1200.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 1200.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 720*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 360*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 19200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^3*m_λ^2) + 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[5, 1, 0, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(640*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 128*δ*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 128*δ*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^7) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(128*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 128*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 128*δ*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 33.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(128*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 128*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 128*δ*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 100.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 60*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^2 - 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 133.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2/m_λ^2 - 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) - 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^2/m_λ^2 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^2/m_λ^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^7 - 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^6 - 160*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 80*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 640*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^6 + 8533.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) - 8533.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*dabs_δ^2*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^5*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 80*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 120*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 40*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 120*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^3*m_λ^2) + 266.666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 533.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^3*m_λ^2) - 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 - 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 400.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 120*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[4, 1, 0, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(128*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 128*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 128*δ*p_amp*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 6*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 12*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 12*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^2/m_λ^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 2048*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^6 + 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 + 8*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 16*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 - 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^3*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 80.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 48*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 24*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 24*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 1280.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^3*m_λ^2) + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[3, 1, 0, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-96*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 32*δ*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 32*δ*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^2/m_λ^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^5*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^5 + 12*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^3*m_λ^2) - 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 12*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[2, 1, 0, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-16*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 16*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 32*δ*p_amp*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 4*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 26.6666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 53.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^3*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[1, 1, 0, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(8*p_amp*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 8*δ*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*δ*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^3 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[0, 1, 0, 1]
        func = 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^3*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[6, 0, 0, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(128*p_amp*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 128*p_amp*π^6*cos(δ*π/mp_P)^2/mp_P^6)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 45*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*dabs_δ^2/m_λ^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 4*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 300.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^2 + 300.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2/m_λ^2 + 180*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 200.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 13312*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^5*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^6 - 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 960*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 768*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 - 25600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 960*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2)
    end

    if dorder_view==[5, 0, 0, 1]
        func = 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 100.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^2 + 30*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 133.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^5 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 160*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 1066.66666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 2133.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^3*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[4, 0, 0, 1]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-32*p_amp*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 32*p_amp*π^4*cos(δ*π/mp_P)^2/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 20.0*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^2 + 12*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 4*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 512*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 + 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[3, 0, 0, 1]
        func = 3*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*(8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^3*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^3 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P*m_λ^2) - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[2, 0, 0, 1]
        func = (8*p_amp*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 8*p_amp*π^2*cos(δ*π/mp_P)^2/mp_P^2)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 13.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*dabs_δ^2*sin(δ*π/mp_P)^2/m_λ^2 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[1, 0, 0, 1]
        func = -4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P
    end

    if dorder_view==[0, 0, 0, 1]
        func = -4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2
    end

    if dorder_view==[6, 0, 2, 0]
        func = (64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 300.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^4 + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 3000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^4/m_λ^6 - 2683.28157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^3/m_λ^5 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2/m_λ^4 + 3000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^4/m_λ^6 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 900.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ^2/m_λ^4 + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 5.0*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 4.47213595499958*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 4.47213595499958*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 250.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*dabs_δ^2/m_λ^6 - 670.820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2/m_λ^5 + 150.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 2000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^3/m_λ^6 + 250.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2/m_λ^6 - 670.820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2/m_λ^5 + 30.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 150.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 - 8049.84471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^5 + 1800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^4 + 750.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ^2*dabs_δ^2/m_λ^6 - 2012.46117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*dabs_δ^2/m_λ^5 + 450.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 75.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*abs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*abs_δ*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 300.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 150.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 75.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 75.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 60.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 6000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^6 + 1500.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^6 - 4024.92235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^5 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 75.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 1800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 900.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 900.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 450.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 225.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 300.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^6*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 1800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 900.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 42932.505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^5) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*abs_δ^2*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^4) - 11448.6680447989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*abs_δ*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^3) + 11448.6680447989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*abs_δ*π^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^2) - 1920.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^4) + 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^3) + 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^3) - 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) - 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 32000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^6) + 9600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 19200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 3840.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^4) - 1920.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^4) + 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^3) - 1717.30020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 16000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^6) - 42932.505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^5) + 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 9600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 19200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 9600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 9600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2)
    end

    if dorder_view==[5, 0, 2, 0]
        func = 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 1000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^4/m_λ^6 - 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^3/m_λ^5 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2/m_λ^4 - 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^3/m_λ^5 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 25.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 22.3606797749979*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 22.3606797749979*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 22.3606797749979*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 22.3606797749979*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 1000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^3/m_λ^6 + 166.666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2/m_λ^6 - 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2/m_λ^5 + 25.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 1000.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*dabs_δ^3/m_λ^6 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^4 + 75.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 50.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 25.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 22.3606797749979*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 22.3606797749979*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 500.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^6 - 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^5 + 150.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 150.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 75.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^4) + 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^3) - 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 2666.66666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^6) - 7155.41752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^5) + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4)
    end

    if dorder_view==[4, 0, 2, 0]
        func = (-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^4/m_λ^6 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) - 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^5 + 120.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^4 + 120.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2/m_λ^4 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 5.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 4.47213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 4.47213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^6 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^6 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^5 + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 120.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 60.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*dabs_δ^2/m_λ^6 - 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*dabs_δ^2/m_λ^5 + 60.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 15.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 120.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 120.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 60.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 30.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 120.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) - 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 640.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 286.216701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2)
    end

    if dorder_view==[3, 0, 2, 0]
        func = -134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^3/m_λ^5 + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^3/m_λ^6 + 30.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 60.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^4 + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^6 - 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^5 + 15.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 30.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 15.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 60.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 30.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 30.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 15.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) - 71.5541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 71.5541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2)
    end

    if dorder_view==[2, 0, 2, 0]
        func = 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^4 + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 16.6666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*dabs_δ^2/m_λ^6 - 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2/m_λ^5 + 10.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 10.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4
    end

    if dorder_view==[1, 0, 2, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(16.6666666666667*abs_δ*dabs_δ/m_λ^4 + 4.47213595499958*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*dabs_δ/m_λ^2 + 10.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2
    end

    if dorder_view==[0, 0, 2, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(6.66666666666667*abs_δ^2/m_λ^4 + 4.47213595499958*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^3) + 5.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^4 - 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^3 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)/m_λ^2
    end

    if dorder_view==[6, 1, 1, 0]
        func = 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-384*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 384*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7 - 256*δ*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^8)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^3*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^4 - 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^3 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 111.80339887499*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^4 + 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 + 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^4 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^3 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^3 - 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^3 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-384*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 384*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7 - 256*δ*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^8)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 - 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^3 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 33.5410196624969*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^4 + 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 + 670.820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^4 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 670.820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^4 - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^3 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 100.623058987491*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^6*π^7*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^8 - 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 4*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 28621.6701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*dabs_δ^3*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^4) + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ*p_amp^2*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 400.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 858.650103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 7155.41752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 22897.3360895978*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^6*abs_δ*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/(mp_P^8*m_λ^2) - 3434.60041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^6*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^7*m_λ^2) - 3434.60041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^7*m_λ^2) - 25600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^3) + 1788.85438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8.94427190999916*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 447.213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 268.328157299975*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 3434.60041343968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^6*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^7*m_λ^2) + 28621.6701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^4) + 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 17173.0020671984*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*abs_δ*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 5366.5631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) - 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*abs_δ*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ*p_amp^2*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 268.328157299975*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 268.328157299975*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 134.164078649987*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8586.50103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 2683.28157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 1609.96894379985*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 804.984471899924*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[5, 1, 1, 0]
        func = 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7) + 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 223.606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^3/m_λ^4 - 66.6666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^3 + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 74.535599249993*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^4 + 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^3 - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^3 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^3 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 11.1803398874989*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 223.606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 + 223.606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^4 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 223.606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 4266.66666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^3) + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 357.770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^7*m_λ^2) + 4770.27835199955*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^4) + 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 894.427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) - 266.666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 894.427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 44.7213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 44.7213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 2862.16701119973*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 298.142396999972*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 178.885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 1431.08350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*abs_δ*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 44.7213595499958*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 894.427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 536.65631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 268.328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[4, 1, 1, 0]
        func = 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 - 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^3 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^4 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^3 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 4*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 143.108350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 357.770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) - 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) - 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8.94427190999916*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 572.433402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 178.885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 178.885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 107.33126291999*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[3, 1, 1, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^3 + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^4 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 143.108350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*abs_δ*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 53.665631459995*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[2, 1, 1, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 26.6666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^3) + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 29.8142396999972*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^4) + 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 17.8885438199983*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[1, 1, 1, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3) + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[0, 1, 1, 0]
        func = 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*abs_δ*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[6, 0, 1, 0]
        func = (64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 100.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^3 + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 447.213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^3/m_λ^4 - 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2/m_λ^3 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ^2/m_λ^3 + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 2.23606797749979*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 111.80339887499*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2/m_λ^4 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 111.80339887499*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2/m_λ^4 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 1341.64078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^4 - 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^3 + 335.410196624968*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*dabs_δ^2/m_λ^4 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 670.820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^4 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 402.492235949962*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 100.623058987491*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 201.246117974981*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^6*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 7155.41752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 5724.33402239946*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*abs_δ*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^2) - 858.650103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) - 858.650103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) - 6400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 858.650103359919*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 7155.41752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 4293.2505167996*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2146.6252583998*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2)
    end

    if dorder_view==[5, 0, 1, 0]
        func = 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 223.606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^3/m_λ^4 - 66.6666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2/m_λ^3 + 223.606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^3/m_λ^4 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 11.1803398874989*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 11.1803398874989*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 74.535599249993*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2/m_λ^4 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 11.1803398874989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^3 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 11.1803398874989*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 223.606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^4 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 134.164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 67.0820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 33.5410196624969*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 1066.66666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^3) + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 143.108350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^5*m_λ^2) + 1192.56958799989*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^4) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 357.770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 715.541752799933*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 357.770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2)
    end

    if dorder_view==[4, 0, 1, 0]
        func = (-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 89.4427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^3/m_λ^4 - 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^3 - 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2/m_λ^3 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 2.23606797749979*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^4 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 8.94427190999916*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 44.7213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*dabs_δ^2/m_λ^4 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 26.8328157299975*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 143.108350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 143.108350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 143.108350559987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2)
    end

    if dorder_view==[3, 0, 1, 0]
        func = 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^3/m_λ^4 + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) - 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^3 + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 22.3606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^4 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 13.4164078649987*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 6.70820393249937*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 35.7770876399966*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*abs_δ*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2)
    end

    if dorder_view==[2, 0, 1, 0]
        func = -6.66666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^3 + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 7.4535599249993*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*dabs_δ^2/m_λ^4 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 4.47213595499958*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2
    end

    if dorder_view==[1, 0, 1, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-5.0*abs_δ*dabs_δ/m_λ^3 - 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2 + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2
    end

    if dorder_view==[0, 0, 1, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1.66666666666667*abs_δ^2/m_λ^3 - 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ^2) + 2.23606797749979*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*abs_δ*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)/m_λ^2
    end

    if dorder_view==[6, 2, 0, 0]
        func = 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)^2 + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1920*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7 + 768*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 768*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 + 256*δ^2*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^9)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(2688*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 2688*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 256*δ^2*p_amp^2*π^8*sin(δ*π/mp_P)^2/mp_P^10 + 256*δ^2*p_amp^2*π^8*cos(δ*π/mp_P)^2/mp_P^10 + 3584*δ*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^9)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(-320*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 320*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 64*δ^2*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 64*δ^2*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 640*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)^2*dabs_δ^2/m_λ^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-320*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 320*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 64*δ^2*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 64*δ^2*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 640*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-320*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 320*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 64*δ^2*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 64*δ^2*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 640*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-320*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 320*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 64*δ^2*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 64*δ^2*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 640*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1920*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7 + 768*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 768*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 + 256*δ^2*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^9)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)/m_λ^2 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 20480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^6*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^9 + 3072*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^8 - 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^8*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^4/mp_P^10 + 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^8*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^2/mp_P^10 + 40960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^8*π^8*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^4*cos(δ*π/mp_P)^4/mp_P^10 + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^7*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^9 - 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^7*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^9 - 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^7*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^9 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 8*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 200.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 200.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 3072*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^8 - 3072*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*(-384*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 384*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7 - 256*δ*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^8)*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^7*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^9 - 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^7*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^9 - 6144*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^7*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^9 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^6*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ^2*p_amp^4*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 400.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 800.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 480*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 25600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^2) + 25600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/(mp_P^2*m_λ^2) + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P^2 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P^2 + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/(mp_P^7*m_λ^2) + 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^7*m_λ^2) + 51200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/(mp_P^7*m_λ^2) + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 + 7680*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P^2 + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 15360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^6*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^4*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 240*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 2400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 1440*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 720*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[5, 2, 0, 0]
        func = 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-320*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 320*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 64*δ^2*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 64*δ^2*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 640*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-1920*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7 + 768*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 768*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 + 256*δ^2*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^9)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 33.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)/m_λ^2 + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-320*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 320*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 64*δ^2*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 64*δ^2*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 640*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2/m_λ^2 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 512*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^8 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^9 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^9 - 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^9 + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 80*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 4266.66666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^6*m_λ^2) - 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P^2 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 2133.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/(mp_P^7*m_λ^2) + 2133.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/(mp_P^7*m_λ^2) + 8533.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*dabs_δ^2*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/(mp_P^7*m_λ^2) - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 5120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 133.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 133.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^6*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^4*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 40*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 80*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 266.666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 266.666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P^2 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 533.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 80*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) - 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 480*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 1600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2)
    end

    if dorder_view==[4, 2, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-320*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 320*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 64*δ^2*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^8 - 64*δ^2*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^8 - 640*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2/m_λ^2 + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 512*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 8*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 512*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 + 512*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 - 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/mp_P^2 + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 1024*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ^2*p_amp^4*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 192*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 320.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 96*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[3, 2, 0, 0]
        func = 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(192*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 128*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 128*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 - 64*δ^2*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5)/m_λ^2 + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) - 128*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)^3/mp_P^7 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)/mp_P^7 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^6*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^7 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 160.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4
    end

    if dorder_view==[2, 2, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(24*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 24*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 - 16*δ^2*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^6 + 16*δ^2*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^6 + 96*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 13.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*dabs_δ^2*sin(δ*π/mp_P)^2/(mp_P^4*m_λ^2) - 13.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*dabs_δ^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 26.6666666666667*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) - 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 53.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^4*m_λ^2) + 32*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ^2*p_amp^4*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[1, 2, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-8*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^4 - 16*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^4 + 16*δ^2*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4
    end

    if dorder_view==[0, 2, 0, 0]
        func = 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2/mp_P^4 - 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^2*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*cos(δ*π/mp_P)^2/mp_P^4 - 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ^2*p_amp^4*π^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^4
    end

    if dorder_view==[6, 1, 0, 0]
        func = 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-384*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 384*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7 - 256*δ*p_amp^2*π^7*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^8)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6)/m_λ^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^2 + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 10240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^6*π^7*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^3*cos(δ*π/mp_P)^3/mp_P^8 - 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 4*(64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 1920*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 1536*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 200.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 120*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 12800.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 3840*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 240*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 600.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 360*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 180*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 1200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[5, 1, 0, 0]
        func = 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(320*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6 - 64*δ*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^7 + 64*δ*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^7)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 33.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5)/m_λ^2 + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^2 + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 - 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^7 + 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 2133.33333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*dabs_δ^2*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/(mp_P^5*m_λ^2) + 1280*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 20*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 640*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 20*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 133.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 120*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 40*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 400.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2)
    end

    if dorder_view==[4, 1, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(64*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 64*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5 + 64*δ*p_amp^2*π^5*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^6) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*dabs_δ^2/m_λ^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 4*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 256*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 80.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 48*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 24*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[3, 1, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*(-48*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4 + 16*δ*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^5 - 16*δ*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^5) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)/m_λ^2 + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^4*π^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^5 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 40.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[2, 1, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-8*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 8*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3 - 16*δ*p_amp^2*π^3*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^4)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 13.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^2*m_λ^2) + 8*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[1, 1, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 - 4*δ*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^3 + 4*δ*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^3)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2 + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[0, 1, 0, 0]
        func = 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*δ*p_amp^2*π*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^2
    end

    if dorder_view==[6, 0, 0, 0]
        func = (64*p_amp^2*π^6*sin(δ*π/mp_P)^2/mp_P^6 - 64*p_amp^2*π^6*cos(δ*π/mp_P)^2/mp_P^6)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 50.0*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^2 + 30*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 50.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*dabs_δ^2/m_λ^2 + 60*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 150.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*dabs_δ^2/m_λ^2 + 90*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 45*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 300.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^2 + 2560*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^4*π^6*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)^2*cos(δ*π/mp_P)^2/mp_P^6 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 - 384*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 3200.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 960*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3
    end

    if dorder_view==[5, 0, 0, 0]
        func = 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 5*(-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 33.3333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*dabs_δ^2/m_λ^2 + 30*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 15*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 100.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^2 - 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^5*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^5 + 533.333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*dabs_δ^2*sin(δ*π/mp_P)*cos(δ*π/mp_P)/(mp_P^3*m_λ^2) + 320*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 160*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3
    end

    if dorder_view==[4, 0, 0, 0]
        func = (-16*p_amp^2*π^4*sin(δ*π/mp_P)^2/mp_P^4 + 16*p_amp^2*π^4*cos(δ*π/mp_P)^2/mp_P^4)*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 4*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^4*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*dabs_δ^2/m_λ^2 + 20.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*dabs_δ^2/m_λ^2 + 12*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 6*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3 + 64*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3
    end

    if dorder_view==[3, 0, 0, 0]
        func = 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 10.0*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*dabs_δ^2/m_λ^2 + 3*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + 16*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*p_amp^2*π^3*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P^3
    end

    if dorder_view==[2, 0, 0, 0]
        func = 3.33333333333333*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*dabs_δ^2/m_λ^2 + 2*exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)^2*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(4*p_amp^2*π^2*sin(δ*π/mp_P)^2/mp_P^2 - 4*p_amp^2*π^2*cos(δ*π/mp_P)^2/mp_P^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)
    end

    if dorder_view==[1, 0, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1.66666666666667*abs_δ*dabs_δ/m_λ^2 + 2.23606797749979*dabs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ) + exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(-2.23606797749979*dabs_δ/m_λ - 4*p_amp^2*π*sin(δ*π/mp_P)*cos(δ*π/mp_P)/mp_P)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)
    end

    if dorder_view==[0, 0, 0, 0]
        func = exp(-2.23606797749979*abs_δ/m_λ - 2*p_amp^2*sin(δ*π/mp_P)^2)*(1 + 2.23606797749979*abs_δ*(1 + 0.74535599249993*abs_δ/m_λ)/m_λ)
    end

    dorder[2] = dorder2  # resetting dorder[2]
    return powers_of_negative_one(dorder2) * float(func)  # correcting for amount of t2 derivatives

end


return mp_kernel, 3  # the function handle and the number of kernel hyperparameters
